module: reflection-internals
synopsis: Enhanced reflection capabilities.
author: Matthias Hlzl
copyright: (c) 1999, Matthias Hlzl

// Returns all subclasses of class that we can know about.  If class
// or one of its subclasses is sealed this may not be all subclasses
// of class.
//
define method all-subclasses (class :: <class>)
 => (subclasses :: <sequence>, unknown-subclasses :: <boolean>);
  // format-out("all-subclasses{<class>}: %=", class);
  *standard-output*.force-output;
  block ()
    all-subclasses(direct-subclasses(class));
  exception (<sealed-object-error>)
    values(#(), #t);
  end block;
end method all-subclasses;

define method all-subclasses (classes :: <sequence>)
  => (subclasses :: <sequence>, unknown-subclasses :: <boolean>);
  // format-out("all-subclasses{<sequence>}: %=\n", classes);
  // *standard-output*.force-output;
  let unknown-subclasses :: <boolean> = #f;
  let subclasses :: <sequence> = #[];
  for (c in classes)
    let new-subclasses = calculate-new-subclasses(c, classes);
    if (new-subclasses)
      subclasses := union(subclasses, new-subclasses);
    else
      unknown-subclasses := #t;
    end if;
  end for;
  if (subclasses.empty?)
    values(classes, unknown-subclasses);
  else
    all-subclasses(union(classes, subclasses));
  end if;
end method all-subclasses;

define function calculate-new-subclasses
    (class :: <class>, known-subclasses :: <sequence>)
 => new-subclasses :: false-or(<sequence>);
  block ()
    let subclasses = direct-subclasses(class);
    for (duplicate in known-subclasses)
      // format-out("  removing %=\n", duplicate);
      subclasses := remove(subclasses, duplicate);
    end for;
    // format-out("  adding %=\n", subclasses);
    subclasses;
  exception (<sealed-object-error>)
    #f;
  end block;
end function calculate-new-subclasses;
